@@ -10,6 +10,10 @@ class TestConvertQuery(unittest.TestCase):
     def test_simple(self):
         self.assertEqual(self._call("stuff"), "'stuff'")
 
+    def test_multiple_words(self):
+        self.assertEqual(self._call("stuff I like"),
+            "( 'stuff' ) & ( 'I' ) & ( 'like' )")
+
     def test_extract_from_dict(self):
         self.assertEqual(self._call({'query': "stuff"}), "'stuff'")
 
@@ -35,6 +39,12 @@ class TestConvertQuery(unittest.TestCase):
     def test_glob(self):
         self.assertEqual(self._call('stuff*'), "'stuff':*")
 
+    def test_word_with_backslash(self):
+        self.assertEqual(self._call("more\\"), "'more\\\\'")
+
+    def test_word_with_apostrophe(self):
+        self.assertEqual(self._call("O'Malley"), "'O''Malley'")
+
 
 def test_suite():
     return unittest.TestSuite((
